/*
 * $Id$
 *
 * by Cedric Duval <cedricduval@free.fr>
 *
 * Copyright (C) Cedric Duval
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif
#include "gettext.h"
#include "abook.h"
#include "db_access.h"
#include "database.h"
#include "misc.h"
#include "options.h"
#include "xmalloc.h"


extern list_item *database;
extern int items;

extern abook_field_list *fields_list;
#define ITEM_SIZE (fields_list->count * sizeof(char *))

extern abook_field standard_fields[];
extern int standard_fields_indexed[];

extern char *selected;




/*
 * Traversal operations.
 * ---------------------
 */

int
real_db_enumerate_items(struct db_enumerator e)
{
	int item = max(0, e.item + 1);
	int i;
	switch(e.mode) {
#ifdef DEBUG
		case ENUM_ALL:
			break;
#endif
		case ENUM_SELECTED:
			for(i = item; i <= last_item(); i++) {
				if(is_selected(i)) {
					item = i;
					goto out;
				}
			}
			return -1;
#ifdef DEBUG
		default:
			fprintf(stderr, "real_db_enumerate_items() "
					"BUG: unknown db_enumerator mode: %d\n",
					e.mode);
			break;
#endif
	}
out:
	return (item > last_item() || item < 0) ? -1 : item;
}

struct db_enumerator
init_db_enumerator(int mode)
{
	struct db_enumerator e;

	e.item = -1; /* important - means "start from beginning" */
	e.mode = mode;

	return e;
}

int
real_enumerate_fields(struct field_enumerator *e)
{
	int i = e->index + 1, j;

	switch(e->mode) {
		case ENUM_EMAILS:
			for(; i < fields_list->count; i++)
				if(fields_list->fields[i]->type == FIELD_EMAILS)
					goto out_normal;
			return e->index = -1;
		case ENUM_STANDARD:
			for(j = e->std_id + 1; standard_fields[j].key; j++)
				if(standard_fields_indexed[j] != -1)
					goto out_std_id;
			return e->std_id = e->index = -1;
	}
out_normal:
	return e->index = (i < fields_list->count ? i : -1);
out_std_id:
	e->std_id = j;
	e->index = standard_fields_indexed[j];
	return e->index;
}

struct field_enumerator
init_field_enumerator(int mode)
{
	struct field_enumerator e;

	e.index = -1; /* important - means "start from beginning" */
	e.std_id = -1;
#ifdef DEBUG
	switch(mode) {
		case ENUM_ALL:
		case ENUM_STANDARD:
		case ENUM_EMAILS:
#endif
			e.mode = mode;
#ifdef DEBUG
			break;
		default:
			assert(0);
	}
#endif

	return e;
}


/*
 * Database commit.
 * ----------------
 */

static void
validate_item(list_item item)
{
	int i, max_field_len;
	char *tmp;

	for(i = 0; i < fields_list->count; i++) {
		max_field_len = 0;

		switch(fields_list->fields[i]->type) {
			case FIELD_EMAILS:
				max_field_len = MAX_EMAILSTR_LEN;
				if(item[i] == NULL)
					item[i] = xstrdup("");
				break;
			case FIELD_LIST:
				/* TODO quote string if it contains commas */
				break;
			case FIELD_STRING:
				max_field_len = MAX_FIELD_LEN;
				break;
			case FIELD_DAY:
				break;
			default:
				assert(0);
		}

		if(max_field_len && item[i] &&
				((int)strlen(item[i]) > max_field_len)) {
			/* truncate field */
			tmp = item[i];
			item[i][max_field_len - 1] = 0;
			item[i] = xstrdup(item[i]);
			free(tmp);
		}
	}
}


int
real_add_item2database(list_item item)
{
	/* 'name' field is mandatory */
	if((item[field_id(NAME)] == NULL) || ! *item[field_id(NAME)]) {
		item_empty(item);
		return 1;
	}

	items++;
	adjust_list_capacity();

	validate_item(item);

	selected[last_item()] = 0;

	database[last_item()] = item_create();
        item_copy(database[last_item()], item);

	return 0;
}


/*
 * Item manipulations.
 * -------------------
 */

list_item item_do(enum item_ops op,
		list_item dest, list_item src, list_item *item)
{
	int i;

	switch(op) {
		case CREATE: return xmalloc0(ITEM_SIZE);
		case RESIZE: *item = xrealloc(*item, ITEM_SIZE); break;
		case FREE: assert(item); xfree(*item); break;
		case EMPTY:
			assert(dest);
			for(i = 0; i < fields_list->count; i++)
				if(dest[i])
					xfree(dest[i]);
			break;
		case BLANK: memset(dest, 0, ITEM_SIZE);	break;
		case COPY: memmove(dest, src, ITEM_SIZE); break;
		case DUPLICATE:
			for(i = 0; i < fields_list->count; i++)
				dest[i] = src[i] ? xstrdup(src[i]) : NULL;
			break;
		default:
			assert(0);
	}
	return NULL;
}


int
real_item_fput(list_item item, int i, char *val)
{
	int id = field_id(i);

	if(id != -1) {
		item[id] = val;
		return 1;
	}

	return 0;
}

char *
real_item_fget(list_item item, int i)
{
	int id = field_id(i);

	if(id != -1)
		return item[id];
	else
		return NULL;
}


/*
 * Database read operations.
 * -------------------------
 */

char *
real_db_field_get(int item, int i, int std)
{
	int id;

	assert(database[item]);

	id = std ? field_id(i) : i;

	if(id != -1)
		return database[item][id];
	else
		return NULL;
}

char *
real_get_first_email(int item)
{
	char *comma, *str = db_email_get(item);

	comma = strchr(str, ',');
	return str ?
		(comma ? xstrndup(str, comma - str) : xstrdup(str)) :
			xstrdup("");
}

int
db_n_items()
{
	return items;
}


/*
 * Database write operations.
 * --------------------------
 */

int
real_db_field_put(int item, int i, int std, char *val)
{
	int id;

	assert(database[item]);

	id = std ? field_id(i) : i;

	if(id != -1) {
		database[item][id] = val;
		return 1;
	}

	return 0;
}

/* 
 * Things like item[field_id(NICK)] should never be used, since besides NAME
 * and EMAIL, none of the standard fields can be assumed to be existing.
 *
 * Prefer the functions item_fput(), item_fget(), db_fput() and db_fget()
 * to access fields in items and database.
 */

/*
 * Field operations.
 * -----------------
 */

/* quick lookup by "standard" field number */
int
real_field_id(int i)
{
	assert((i >= 0) && (i < ITEM_FIELDS));
	return standard_fields_indexed[i];
}

union ret_value
field_query(enum field_ops action, int i, char *s)
{
	union ret_value res;

	switch(action) {
		case QUERY_KEY:	res.str_v = fields_list->fields[i]->key; break;
		case QUERY_NAME: res.str_v = fields_list->fields[i]->name;break;
		case QUERY_TYPE: res.int_v = fields_list->fields[i]->type;break;
		case QUERY_INDEX:
			res.int_v = -1;

			if(!s)
				break;

			for(i = 0; i < fields_list->count; i++)
				if( !strcmp(fields_list->fields[i]->key, s) ) {
					res.int_v = i;
					goto out;
				}
			break;
		default:
			assert(0);
	}
out:
	return res;
}

/*
 * Declare a new field while database is already loaded.
 * 
 * Returns the index of the new last item, or -1 if the field was already
 * declared.
 */
int
real_declare_unknown_field(char *key)
{
	int i;

	if(declare_new_field(key, key, "string",
			1 /* accept to declare "standard" fields */) != NULL)
		return -1;

	if(!database)
		goto out;

	for(i = 0; i < items; i++)
		if(database[i]) {
			database[i] = xrealloc(database[i], ITEM_SIZE);
			database[i][fields_list->count - 1] = NULL;
		}
out:
	return fields_list->count - 1;
}

