
/*
 * $Id: filter.c,v 1.50 2005/10/26 19:29:16 jheinonen Exp $
 *
 * by JH <jheinonen@users.sourceforge.net>
 *
 * Copyright (C) Jaakko Heinonen
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ltdl.h>
#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif
#include "abook_curses.h"
#include "abook.h"
#include "db_access.h"
#include "filter.h"
#include "gettext.h"
#include "list.h"
#include "misc.h"
#include "ui.h"
#include "xmalloc.h"
#include <assert.h>


/*
 * common functions
 */

static void
readonly_ops(struct abook_ops *ops)
{
	memset(ops, 0, sizeof(struct abook_ops));

	ops->real_db_field_get = real_db_field_get;
	ops->real_get_first_email = real_get_first_email;
	ops->real_db_enumerate_items = real_db_enumerate_items;
	ops->real_enumerate_fields = real_enumerate_fields;
	ops->real_field_id = real_field_id;
	ops->field_query = field_query;
	ops->opt_get_value = opt_get_value;
}

static void
readwrite_ops(struct abook_ops *ops)
{
	readonly_ops(ops);

	ops->real_db_field_put = real_db_field_put;
	ops->real_add_item2database = real_add_item2database;
	ops->item_do = item_do;
	ops->real_item_fget = real_item_fget;
	ops->real_item_fput = real_item_fput;
	ops->real_declare_unknown_field = real_declare_unknown_field;
}

/* given a filter name, get the functions and the associated handle */
static struct filter_ops *
open_filter(char *name, lt_dlhandle *handle)
{
	init_func_t *filter_init;
	struct filter_ops *ops;
	struct abook_ops my_ops;

	if(lt_dlinit())
		return NULL;

	lt_dladdsearchdir(PATH_FILTERS_IN_BUILD_TREE);
	lt_dladdsearchdir(FILTERS_DIR);

	if(!(*handle = lt_dlopenext(name)))
		return NULL;

	if (!(filter_init = lt_dlsym(*handle, "filter_init"))) {
		lt_dlclose(*handle);
		return NULL;
	}

	readwrite_ops(&my_ops);
	ops = (*filter_init)(&my_ops);

	if(!ops)
		lt_dlclose(*handle);

	return ops;
}

struct filters_info {
	abook_list *input;
	abook_list *output;
};

static int
get_filter_info(const char *path, lt_ptr data)
{
	lt_dlhandle handle;
	struct filter_ops *ops;
	struct filters_info *infos = (struct filters_info *)data;
	char *str, *p, *q, *name;

	if(!(p = strstr(path, "_filter")))
		return 0;

	if(!(ops = open_filter((char *)path, &handle))) {
		lt_dlexit();
		return 0;
	}

	q = strrchr(path, '/');
	q = q ? q + 1 : (char *)path;
	name = xstrndup(q, p - q);

	str = strdup_printf("%s%c%s", name,
			ops->description ? '\t' : '\0',
			ops->description ? safe_str(ops->description()) : "");
	free(name);

	if(ops->import)
		abook_list_append(&infos->input, str);

	if(ops->export)
		abook_list_append(&infos->output, str);

	free(str);
	lt_dlclose(handle);
	lt_dlexit();

	return 0;
}

static int
filters_info(struct filters_info *parms)
{
	struct stat s;

	if(lt_dlinit ())
		return -1;

	if(!lstat(PATH_FILTERS_IN_BUILD_TREE, &s) && S_ISDIR(s.st_mode))
		lt_dlforeachfile(PATH_FILTERS_IN_BUILD_TREE,
				get_filter_info, (lt_ptr)parms);
	else
		lt_dlforeachfile(FILTERS_DIR, get_filter_info, (lt_ptr)parms);

	lt_dlexit();

	return 0;
}

void
print_filters()
{
	abook_list *cur;
	struct filters_info infos = { NULL, NULL };

	filters_info(&infos);

	puts(_("input:"));
	for(cur = infos.input; cur; cur = cur->next)
		printf("\t%s\n", cur->data);

	putchar('\n');

	puts(_("output:"));
	for(cur = infos.output; cur; cur = cur->next)
		printf("\t%s\n", cur->data);

	abook_list_free(&infos.input);
	abook_list_free(&infos.output);
}

static char *
filters_screen(abook_list *list, enum filter_actions action)
{
	int i, filter;
	char *str, *p;
	abook_list *cur;

	clear();

	refresh_statusline();
	switch(action) {
		case FILTER_IMPORT: headerline(_("import database")); break;
		case FILTER_EXPORT: headerline(_("export database")); break;
		default: break;
	}

	mvaddstr(3, 1, _("please select a filter"));

	for(cur = list, i = 0; cur; cur = cur->next, i++)
		mvprintw(5 + i, 6, "%c -\t%s\n", 'a' + i,
			cur->data);

	mvprintw(6 + i, 6, _("x -\tcancel"));

	do {
		if((filter = getch() - 'a') == ('x' - 'a')) {
			refresh_screen();
			abook_list_free(&list);
			return NULL;
		}
	} while(filter >= abook_list_count(list) && filter < 0);

	mvaddstr(5 + filter, 2, "->");

	str = p = abook_list_get(list, filter);
	SKIPNWS(p);
	str = xstrndup(str, p - str);

	return str;
}

/*
 * end of common functions
 */


/*
 * import
 */

static int		i_read_file(char *filename, int (*func) (FILE *in));

static import_func_t *
import_func(char *filtname, lt_dlhandle *handle)
{
	struct filter_ops *ops;
	char *name = strdup_printf("%s_filter", filtname);

	ops = open_filter(name, handle);
	free(name);
	if(!ops || !ops->import) {
		lt_dlexit();
		return NULL;
	}

	return ops->import;
}
	
void
import_database()
{
	char *filename, *filtname;
	int tmp = db_n_items();
	struct filters_info infos = { NULL, NULL };
	lt_dlhandle handle;
	import_func_t *func;

	filters_info(&infos);
	abook_list_free(&infos.output);

	filtname = filters_screen(infos.input, FILTER_IMPORT);
	abook_list_free(&infos.input);
	if(!filtname)
		goto out_import1;

	filename = ask_filename(_("Filename: "));
	if(!filename)
		goto out_import1;

	if((func = import_func(filtname, &handle)) == NULL)
		goto out_import2;

	if(i_read_file(filename, func))
		statusline_msg(_("Error occured while opening the file"));
	else if(tmp == db_n_items())
		statusline_msg(_("File does not seem to be a valid addressbook"));

	free(filename);
out_import2:
	lt_dlclose(handle);
	lt_dlexit();
out_import1:
	refresh_screen();
}

static int
i_read_file(char *filename, int (*func) (FILE *in))
{
	FILE *in;
	int ret = 0;

	if( (in = abook_fopen( filename, "r" )) == NULL )
		return 1;

	ret = (*func) (in);

	fclose(in);

	return ret;
}

int
import_file(char filtname[FILTNAME_LEN], char *filename)
{
	int ret = 0, tmp = db_n_items();
	lt_dlhandle handle;
	import_func_t *func;
	
	if((func = import_func(filtname, &handle)) == NULL)
		return 1;
	
	if(!strcmp(filename, "-")) {
		struct stat s;
		if((fstat(fileno(stdin), &s)) == -1 || S_ISDIR(s.st_mode))
			ret = 1;
		else
			ret = (*func) (stdin);
	} else
		ret = i_read_file(filename, func);

	if(tmp == db_n_items())
		ret = 1;

	lt_dlclose(handle);
	lt_dlexit();

	return ret;
}


/*
 * export
 */

static int e_write_file(char *filename,
		int (*func) (FILE *in, struct db_enumerator e), int mode);

static export_func_t *
export_func(char *filtname, lt_dlhandle *handle)
{
	struct filter_ops *ops;
	char *name = strdup_printf("%s_filter", filtname);

	ops = open_filter(name, handle);
	free(name);
	if(!ops || !ops->export) {
		lt_dlexit();
		return NULL;
	}

	return  ops->export;
}

void
export_database()
{
	int enum_mode = ENUM_ALL;
	char *filename, *filtname;
	struct filters_info infos = { NULL, NULL };
	lt_dlhandle handle;
	export_func_t *func;

	filters_info(&infos);
	abook_list_free(&infos.input);

	filtname = filters_screen(infos.output, FILTER_EXPORT);
	abook_list_free(&infos.output);
	if(!filtname)
		goto out_export1;

	if(selected_items()) {
		switch(statusline_askchoice(
			_("Export <a>ll, export <s>elected, or <c>ancel?"),
			S_("keybindings:all/selected/cancel|asc"), 3)) {
			case 1:
				break;
			case 2:
				enum_mode = ENUM_SELECTED;
				break;
			case 0:
			case 3:
				goto out_export1;
		}
		clear_statusline();
	}

	filename = ask_filename(_("Filename: "));
	if(!filename)
		goto out_export1;

	if((func = export_func(filtname, &handle)) == NULL)
		goto out_export2;

	if(e_write_file(filename, func, enum_mode))
		statusline_msg(_("Error occured while exporting"));

	lt_dlclose(handle);
	lt_dlexit();
out_export2:
	free(filename);
out_export1:
	refresh_screen();
}

static int
e_write_file(char *filename, int (*func) (FILE *in, struct db_enumerator e),
		int mode)
{
	FILE *out;
	int ret = 0;
	struct db_enumerator enumerator = init_db_enumerator(mode);

	if((out = fopen(filename, "a")) == NULL)
		return 1;

	if(ftell(out))
		return 1;

	ret = (*func) (out, enumerator);

	fclose(out);

	return ret;
}

int
fexport(char filtname[FILTNAME_LEN], FILE *fp, int enum_mode)
{
	int ret;
	struct db_enumerator e = init_db_enumerator(enum_mode);
	lt_dlhandle handle;
	export_func_t *f;

	if((f = export_func(filtname, &handle)) == NULL)
		return -1;

	ret = (*f) (fp, e);

	lt_dlclose(handle);
	lt_dlexit();

	return ret;
}

int
export_file(char filtname[FILTNAME_LEN], char *filename)
{
	const int mode = ENUM_ALL;
	int ret = 0;
	struct db_enumerator e = init_db_enumerator(mode);
	lt_dlhandle handle;
	export_func_t *f;

	if((f = export_func(filtname, &handle)) == NULL)
		return -1;

	if(!strcmp(filename, "-"))
		ret = (*f) (stdout, e);
	else
		ret =  e_write_file(filename, f, mode);

	lt_dlclose(handle);
	lt_dlexit();

	return ret;
}


